package cs108;

import java.util.Iterator;
import java.util.NoSuchElementException;

public final class SHashSet<E> extends SAbstractSet<E> {
    private static int MIN_CAPACITY = 10;
    private static double MIN_LOAD_FACTOR = 0.4;
    private static double MAX_LOAD_FACTOR = 1;
    private static double IDEAL_LOAD_FACTOR = 0.7;

    private int size;
    private SList<E>[] table;

    public SHashSet() {
        this.size = 0;
        this.table = newTable(MIN_CAPACITY);
    }

    @Override
    public int size() {
        return size;
    }

    @Override
    public void add(E e) {
        SList<E> list = listFor(table, e);
        if (! list.contains(e)) {
            list.add(0, e);
            size += 1;
            rehashIfNeeded();
        }
    }

    @Override
    public void remove(E e) {
        Iterator<E> listIt = listFor(table, e).iterator();
        while (listIt.hasNext()) {
            E e1 = listIt.next();
            if (e1.equals(e)) {
                listIt.remove();
                size -= 1;
                rehashIfNeeded();
                return;
            }
        }
    }

    @Override
    public boolean contains(E e) {
        return listFor(table, e).contains(e);
    }

    @Override
    public Iterator<E> iterator() {
        return new Iterator<E>() {
            int remaining = size;
            int listIndex = 0;
            Iterator<E> listIt = table[listIndex].iterator();

            @Override
            public boolean hasNext() {
                return remaining > 0;
            }

            @Override
            public E next() {
                if (! hasNext())
                    throw new NoSuchElementException();
                remaining -= 1;
                while (! listIt.hasNext()) {
                    listIndex += 1;
                    listIt = table[listIndex].iterator();
                }
                return listIt.next();
            }

            @Override
            public void remove() {
                listIt.remove();
                size -= 1;
            }
        };
    }

    private static <E> SList<E> listFor(SList<E>[] table, E elem) {
        return table[Math.floorMod(elem.hashCode(), table.length)];
    }

    private static <E> SList<E>[] newTable(int capacity) {
        assert capacity > 0;

        @SuppressWarnings("unchecked")
        SList<E>[] table = new SList[capacity];
        for (int i = 0; i < capacity; ++i)
            table[i] = new SLinkedList<>();
        return table;
    }

    private void rehashIfNeeded() {
        double loadFactor = size / (double)table.length;
        if (MIN_LOAD_FACTOR <= loadFactor && loadFactor <= MAX_LOAD_FACTOR)
            return;

        int idealCapacity = (int) (size / IDEAL_LOAD_FACTOR);
        int newCapacity = Math.max(MIN_CAPACITY, idealCapacity);
        if (newCapacity == table.length)
            return;

        SList<E>[] newTable = newTable(newCapacity);
        for (E e: this)
            listFor(newTable, e).add(0, e);

        table = newTable;
    }
}
